@@ -15,7 +15,7 @@ To download the pairs dataset see
 import logging
 import enum
 from dataclasses import dataclass
-from typing import Optional, List, Iterable, Dict, Union, Set, Tuple, TypeAlias
+from typing import Optional, List, Iterable, Dict, Union, Set, Tuple, TypeAlias, Collection
 
 import pandas as pd
 import numpy as np
@@ -59,7 +59,7 @@ HumanReadableTradingPairDescription: TypeAlias = Tuple[ChainId, str, str, str]
 
 
 @dataclass_json
-@dataclass
+@dataclass(slots=True)
 class DEXPair:
     """ Trading pair information for a single pair.
 
@@ -500,6 +500,10 @@ class PandasPairUniverse:
     def get_all_pair_ids(self) -> List[PrimaryKey]:
         return self.df["pair_id"].unique()
 
+    def get_pair_ids_by_exchange(self, exchange_id: PrimaryKey):
+        """Get all pair ids on a specific exchange"""
+        return self.df.loc[self.df["exchange_id"] == exchange_id]["pair_id"]
+
     def get_count(self) -> int:
         """How many trading pairs there are."""
         return len(self.df)
@@ -514,17 +518,27 @@ class PandasPairUniverse:
             return None.
         """
 
+        # First try the cached suing self.pair_map
         if self.pair_map:
-            # TODO: Eliminate non-indexed code path?
+            # Convert any pairs in-fly to DEXPair objects and store them.
+            # We do not initially construct these objects,
+            # as we do not know what pairs a strategy might access.
             data = self.pair_map.get(pair_id)
 
             if data is None:
                 return None
 
-            return DEXPair.from_dict(data)
+            if isinstance(data, DEXPair):
+                return data
 
-        # TODO: Remove
+            # Convert
+            obj = DEXPair.from_dict(data)
+            self.pair_map[pair_id] = obj
+            return obj
 
+        # We did not build this universe with pair index
+        # Not sure why anyone would really want to do this
+        # maybe eliminate this code path altogether in the future
         df = self.df
 
         pairs: pd.DataFrame = df.loc[df["pair_id"] == pair_id]
@@ -726,10 +740,9 @@ class PandasPairUniverse:
         if exchange is None:
             # Try to produce very helpful error message
             if exchange_universe.get_exchange_count() == 1:
-                exchange_slug = exchange_universe.get_single().exchange_slug
-                exchange_message = f"The slug of the only exchange we have is {exchange_slug}."
+                our_exchange_slug = exchange_universe.get_single().exchange_slug
+                exchange_message = f"The slug of the only exchange we have is {our_exchange_slug}."
             else:
-                exchange_slug = None
                 exchange_message = ""
 
             raise NoPairFound(f"The trading universe does not contain data for the exchange {exchange_slug} on chain {chain_id.name}.\n"
@@ -1018,7 +1031,7 @@ class LegacyPairUniverse:
         return filter(lambda p: p.flag_inactive, self.pairs.values())
 
 
-def filter_for_exchanges(pairs: pd.DataFrame, exchanges: List[Exchange]) -> pd.DataFrame:
+def filter_for_exchanges(pairs: pd.DataFrame, exchanges: Collection[Exchange]) -> pd.DataFrame:
     """Filter dataset so that it only contains data for the trading pairs from a certain exchange.
 
     Useful as a preprocess step for creating :py:class:`tradingstrategy.candle.GroupedCandleUniverse`
